# -*- coding: utf-8 -*-
#
# Univention Application Reporting
#
# Copyright (C) 2015 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of the software contained in this package
# as well as the source package itself are made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this package provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use the software under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import urllib2
import ldap
import ldap.filter
from json import dumps as json_encode
import fcntl
import socket
import struct
import uuid

import univention.debug as ud
import univention.admin.uldap
import univention.admin.modules
import univention.config_registry
from univention.config_registry.interfaces import Interfaces
from univention.admin.uexceptions import base as udm_error
from univention.lib.package_manager import PackageManager

univention.admin.modules.update()
ldap_connection, ldap_position = univention.admin.uldap.getMachineConnection()
ucr = univention.config_registry.ConfigRegistry()
ucr.load()
interfaces = Interfaces()

from univention.management.console.modules.appcenter import app_center
APPS = dict((app.id.lower(), app) for app in app_center.Application.all_installed(PackageManager(), only_local=True, localize=False))


def get_ip_address():
	try:
		ifname = interfaces.primary
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		return socket.inet_ntoa(fcntl.ioctl(s.fileno(), 0x8915, struct.pack('256s', ifname[:15]))[20:24])
	except:
		pass


def get_mac_address():
	try:
		ifname = interfaces.primary
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		info = fcntl.ioctl(s.fileno(), 0x8927,  struct.pack('256s', ifname[:15]))
		return ':'.join(['%02X' % ord(char) for char in info[18:24]])
	except:
		try:
			return ':'.join(("%012X" % uuid.getnode())[i:i+2] for i in range(0, 12, 2))
		except:
			pass


class Request(object):

	def __init__(self, url, body, headers):
		self.url = url
		self.body = body
		self.headers = headers
		self.request = urllib2.Request(url, body, headers)

	def __repr__(self):
		return '<Request(%s, %r)>' % (self.url, self.headers)


class ServiceInterface(object):

	apps = ()

	def url(self, app):
		raise NotImplementedError('url')

	def body(self, app):
		keys = ('uuid/system', 'uuid/license', 'ldap/base', 'hostname', 'domainname')
		body = dict((key, ucr.get(key, '')) for key in keys)
		body['mac'] = get_mac_address()
		body['ip'] = get_ip_address()
		body.update(app.body())
		return body

	def encode_body(self, body):
		return json_encode(body)

	def headers(self, app):
		return {
			'Content-Type': 'application/json'
		}

	def application(self, app):
		return Application(app)

	def __init__(self):
		if '*' in self.apps:
			self._apps = [app for app in APPS.values()]
		else:
			self._apps = [APPS.get(app.lower()) for app in self.apps]

		self._apps = [self.application(app) for app in self._apps]

	def requests(self):
		for app in self._apps:
			yield self.build_request(app)

	def build_request(self, app):
		return Request(self.url(app), self.encode_body(self.body(app)), self.headers(app))


class UniventionServer(ServiceInterface):

	def __init__(self):
		if ucr.is_true('appreport/*/report', False):
			self.apps = ('*',)
		else:
			self.apps = [app for app in APPS if ucr.is_true('appreport/%s/report' % (app,), False)]

		super(UniventionServer, self).__init__()

	def url(self, app):
		return 'https://license.univention.de/appreport/%s' % urllib2.quote(app.id)


class Application(object):

	def __init__(self, app):
		self.app = app
		self.id = app.id
		self.object_type = ucr.get('appreport/%s/object_type' % (self.id,), 'users/user')
		self.object_filter = ucr.get('appreport/%s/object_filter' % (self.id,), '')
		self.object_attribute = ucr.get('appreport/%s/object_attribute' % (self.id,), '')
		self.attribute_type = ucr.get('appreport/%s/attribute_type' % (self.id,), '')
		self.attribute_filter = ucr.get('appreport/%s/attribute_filter' % (self.id,), '')

		# backwards compatibility
		groupmember_search = ucr.get('appreport/%s/groupmember_search' % (self.id,), '')
		if groupmember_search:
			self.object_filter = '(|(%s))' % ')('.join(['name=%s' % (ldap.filter.escape_filter_chars(x),) for x in groupmember_search.split(',')])
			self.object_attribute = 'uniqueMember'
			self.object_type = 'groups/group'

	def object_count(self):
		object_ = univention.admin.modules.get(self.object_type)
		try:
			objects = object_.lookup(None, ldap_connection, self.object_filter)
		except AttributeError as exc:  # object_ is None
			ud.debug(ud.MAIN, ud.ERROR, 'Unknown object type? %r=%r: %s' % (self.object_type, object_, exc))
			raise
		except (ldap.LDAPError, udm_error) as exc:
			ud.debug(ud.MAIN, ud.ERROR, 'LDAP lookup failure: %s' % (exc,))
			raise

		if self.object_attribute:
			attrs = []
			for obj in objects:
				attrs.extend(obj.oldattr[self.object_attribute])
			attrs = set(attrs)

			if self.attribute_type:
				attr_object = univention.admin.modules.get(self.attribute_type)
				attr_objects = attr_object.lookup(None, ldap_connection, self.attribute_filter)
				attrs = [dn for dn in attrs if any(dn.lower() == obj.dn.lower() for obj in attr_objects)]

			return len(attrs)

		return len(objects)

	def body(self):
		return {
			'count': self.object_count()
		}
